# vim: expandtab tabstop=4 shiftwidth=4

from .complex_sinusoid import complex_sinusoid
from .data_types import FullDataType
from .decimal_convert_to_base import decimal_convert_to_base, num2base
from .exceptions import DataTypeException, EndiannessException, NumberSpaceException, SignalTypeException, WriteModeException, RoundingException
from .exceptions import DSPFTWException
from .filename_load import filename_load, fnload
from .generate_qam import generate_qam
from .generate_random_qpsk import generate_random_qpsk
from .load_bits import load_bits, loadbits
from .load_data import load_data, loaddata
from .load_signal import load_signal, loadsig
from .save_bits import save_bits, savebits, save_bits_to
from .save_data import save_data, savedata
from .save_signal import save_signal, savesig
from .signal_correlation import sigcorr, signal_correlation
from .signal_types import FullSignalType
from .truncation import true_one, true_zero
from .raised_cosine_filter_generator import raised_cosine_filter_generator, rcfiltgen
from .root_raised_cosine_filter_generator import root_raised_cosine_filter_generator, rrcfiltgen
from .root_raised_cosine import root_raised_cosine, rrcfunct
from .test_qpsk_modulation import test_qpsk_modulation
from .vector_power import vector_power, vecpow



# vim: expandtab tabstop=4 shiftwidth=4

from sys import byteorder

from typing import Any
from numpy import array as nparray
from numpy import fromfile

from .signal_types import FullSignalType

def load_signal(file_name: str, signal_type: str, number_space: str, endianness: str=byteorder, num_samples: int=-1, start_sample: int=0, rounding: Any=0) -> nparray:
    '''
    Loads a signal from a file in float or two's compliment, real or complex format.

    Parameters
    ----------
    file_name:
        The name of the file to load.
    signal_type:
        The size and type of signal samples, such as '8t', '16t', '32t', '32f', '64f'.
    number_space:
        Number space of the file.  Valid values are 'real', 'r', 'complex', 'cplx', 'c'.
    endianness:
        Endianness of the file, such as 'big', 'b', 'little', 'l'.
        Defaults to the system endianness.
    num_samples:
        The number of samples to load from the file.
    start_sample:
        The number of samples to skip before loading.
    rounding:
        Twos-complement rounding: 0=Asymmetric/Exact Zero, 1=True One.

    Returns a numpy array.
    '''
    full_signal_type = FullSignalType(signal_type, number_space, endianness, rounding)
    data = fromfile(
        file_name,
        dtype=full_signal_type.numpy_dtype(),
        offset=full_signal_type.offset(start_sample),
        count=full_signal_type.count(num_samples),
    )
    return full_signal_type.post_load(data)

def loadsig(*args, **kwargs) -> nparray:
    '''
    Alias for load_signal.
    '''
    return load_signal(*args, **kwargs)



# vim: expandtab tabstop=4 shiftwidth=4

from io import BufferedReader
from typing import Iterable
from numpy import array as nparray

def read_to_bools(reader: BufferedReader, num_bytes: int = -1) -> Iterable[bool]:
    for byte in reader.read(num_bytes):
        if byte == 0:
            yield False
        else:
            yield True

def unpack_to_bools(reader: BufferedReader, num_bytes: int = -1) -> Iterable[bool]:
    for byte in reader.read(num_bytes):
        for offset in range(8):
            yield byte & (0x80>>offset) == 0x80 >> offset

def load_packed_bits(file_name: str, count: int=-1, offset: int=0) -> nparray:
    with open(file_name, 'rb') as bit_file:
        byte_offset = offset // 8
        bit_file.seek(byte_offset)
        bit_offset_into_byte = offset % 8
        reader = BufferedReader(bit_file)
        if count == -1:
            bools = unpack_to_bools(reader, count)
        else:
            bools = unpack_to_bools(reader, count//8+2)
        bool_array = nparray(list(bools), dtype=bool)

    if count == -1:
        return bool_array[bit_offset_into_byte:]

    return bool_array[bit_offset_into_byte:bit_offset_into_byte+count]

def load_unpacked_bits(file_name: str, count: int=-1, offset: int=0) -> nparray:
    with open(file_name, 'rb') as bit_file:
        bit_file.seek(offset)
        reader = BufferedReader(bit_file)
        bools = read_to_bools(reader, count)
        bool_array = nparray(list(bools), dtype=bool)

        if count == -1:
            return bool_array

        return bool_array[:count]

def load_bits(file_name: str, count: int=-1, offset: int=0, packed=True) -> nparray:
    '''
    Loads bits from a file int a bool array.

    Parameters
    ----------
    file_name: string
        The name of the file to load.
    count: int
        The number of bits to load.
        Default is all.
    offset: int
        The number of bits to skip before loading.
        Default is zero.
    packed: bool
        Whether the bits are packed into bytes.
        Default is True.

    Returns a numpy bool array.
    '''

    if packed:
        return load_packed_bits(file_name, count, offset)

    return load_unpacked_bits(file_name, count, offset)

def loadbits(*args, **kwargs) -> nparray:
    '''
    Alias for load_bits.
    '''
    return load_bits(*args, **kwargs)



# vim: expandtab tabstop=4 shiftwidth=4

from typing import Any, Tuple
import numpy as np

from .signal_types import FullSignalType
from .rounding import  Rounding
from .writemode import normalize_write_mode

def save_signal(file_name: str, sig: np.array, signal_type: str, number_space: str,
                endianness: str, write_mode: str='w', rounding: Any=0) -> Tuple[int, int]:
    '''
    Function to use numpy package to write PCM data.

    Parameters
    ----------
    file_name:
        The name of the file to write.
    sig:
        numpy.array containing the signal.
    signal_type:
        The size and type of signal samples, such as '8t', '16t', '32t', '32f', '64f'.
    number_space:
        Number space of the file.  Valid values are 'real', 'r', 'complex', 'cplx', 'c'.
    endianness:
        Endianness of the file, such as 'big', 'b', 'little', 'l'.
    write_mode:
        'w' for new or overwrite. 'a' for append.
        Default is w.
    rounding:
        Rounding behavior for two's complement.
        '0' = True Zero (asymmetric), '1' = True One (symmetric)
        Default is 0.

    Returns a tuple of the number of samples written and the number of samples clipped.
    '''

    write_mode = normalize_write_mode(write_mode)
    full_signal_type = FullSignalType(signal_type, number_space, endianness, rounding)

    if len(sig) == 0:
        sig = np.array([])

    if full_signal_type.is_twos_compliment_signal_type():

        if full_signal_type.is_complex():
            # Flatten complex signals to real
            sigflat = np.array([sig.real, sig.imag]).flatten("F")
        else:
            sigflat = np.array(sig).real

        if rounding == Rounding.TRUEZERO.value:
            # Shift number line up
            sigflat += 1 / (2.0*full_signal_type.scale())
        # if self.rounding == Rounding.TRUEONE:
            # sigflat = sigflat # No change

        # Put values from [-1,1) in interval [-2**R/2,2**R/2) and clip others
        # R = two's complement number (e.g. for 8t, R=8)
        sigout, num_clipped = full_signal_type.clip_real_signal(sigflat)
        # Round down to nearest integer
        sigout = np.array(np.floor(sigout), 'int')

        # Modulo 2^8 or 2^16 to create two's complement
        sigout = np.remainder(sigout, 2*full_signal_type.scale())
        # Cast sample as the desired data type
        sigout = np.array(sigout, full_signal_type.numpy_dtype())

    else:
        num_clipped = 0

        if full_signal_type.is_complex():
            # Flatten and cast sample as the desired data type
            sigout = np.array(
                [sig.real, sig.imag],
                full_signal_type.numpy_dtype()
            ).flatten("F")
        else:
            # Cast sample as the desired data type
            sigout = np.array(sig.real, full_signal_type.numpy_dtype())

    with open(file_name, write_mode.value) as outfile:
        sigout.tofile(outfile)

    num_written = len(sigout) // full_signal_type.items_per_sample()

    return num_written, num_clipped

def savesig(*args, **kwargs):
    '''
    Alias for save_signal.
    '''
    return save_signal(*args, **kwargs)



# vim: expandtab tabstop=4 shiftwidth=4

from enum import Enum, auto
from typing import Any, Tuple

import numpy as np

from .endianness import Endianness, normalize_endianness
from .rounding import Rounding, normalize_rounding
from .exceptions import NumberSpaceException, SignalTypeException

class SignalType(Enum):
    ST8T  = 'b'
    ST16T = 'i2'
    ST32T = 'i4'
    ST32F = 'f4'
    ST64F = 'f8'

class NumberSpace(Enum):
    REAL = auto()
    COMPLEX = auto()

def normalize_signal_type(signal_type: str) -> SignalType:
    signal_type = signal_type.strip().lower()

    if signal_type == '8t':
        return SignalType.ST8T

    if signal_type in ('16t', '16tl', '16tr'):
        return SignalType.ST16T

    if signal_type in ('32t', '32tl', '32tr'):
        return SignalType.ST32T

    if signal_type in ('32f', '32fl', '32fr'):
        return SignalType.ST32F

    if signal_type in ('64f', '64fl', '64fr'):
        return SignalType.ST64F

    raise SignalTypeException(f"Unknown signal type {signal_type}")

def normalize_number_space(number_space: str) -> NumberSpace:
    number_space = number_space.strip().lower()

    if number_space in ('c', 'cplx', 'complex'):
        return NumberSpace.COMPLEX

    if number_space in ('r', 'real'):
        return NumberSpace.REAL

    raise NumberSpaceException(f"Unknown number space {number_space}")

class FullSignalType:
    def __init__(self, signal_type: str, number_space: str, endianness: str, rounding: Any=0) -> None:
        self.signal_type: SignalType = normalize_signal_type(signal_type)
        self.number_space: NumberSpace = normalize_number_space(number_space)
        self.endianness: Endianness = normalize_endianness(endianness)
        self.rounding: Rounding = normalize_rounding(rounding)

    def numpy_dtype(self) -> str:
        return self.endianness.value + self.signal_type.value

    def count(self, count: int) -> int:
        if count == -1:
            return -1

        return count * self.items_per_sample()

    def bytes_per_item(self):
        if self.signal_type == SignalType.ST16T:
            return 2

        if self.signal_type == SignalType.ST32T:
            return 4

        if self.signal_type == SignalType.ST32F:
            return 4

        if self.signal_type == SignalType.ST64F:
            return 8

        return 1

    def offset(self, offset: int) -> int:
        return offset * self.items_per_sample() * self.bytes_per_item()

    def post_load(self, data: np.array) -> np.array:
        scaled_signal = data / self.scale()

        if self.is_twos_compliment_signal_type() and self.rounding == Rounding.TRUEONE:
            scaled_signal += 1 / (2*self.scale())

        if self.is_complex():
            return scaled_signal[0::2] + scaled_signal[1::2]*1j

        return scaled_signal

    def clip_signal(self, signal: np.array) -> Tuple[np.array, int]:
        mult = self.scale()

        if self.is_complex():
            sigflat = np.array([mult*signal.real, mult*signal.imag]).flatten('F')
        else:
            sigflat = np.copy(mult*signal)

        # Count how many will be clipped below
        clip_count = len(np.argwhere(sigflat < -1*mult).flatten())
        # Clipping values that are too low
        sigflat = np.where(sigflat < -1*mult, -1*mult, sigflat)

        # Count how many will be clipped above
        clip_count += len(np.argwhere(sigflat>=mult).flatten())
        # Clipping values that are too high
        sigflat = np.where(sigflat >= mult, mult-1, sigflat)

        return sigflat, clip_count

    def clip_real_signal(self, signal: np.array) -> Tuple[np.array, int]:
        mult = self.scale()

        sigflat = np.copy(mult*signal)

        # Count how many will be clipped below
        clip_count = len(np.argwhere(sigflat < -1*mult).flatten())
        # Clipping values that are too low
        sigflat = np.where(sigflat < -1*mult, -1*mult, sigflat)

        # Count how many will be clipped above
        clip_count += len(np.argwhere(sigflat>=mult).flatten())
        # Clipping values that are too high
        sigflat = np.where(sigflat >= mult, mult-1, sigflat)

        return sigflat, clip_count

    def is_complex(self) -> bool:
        if self.number_space == NumberSpace.COMPLEX:
            return True

        return False

    def is_twos_compliment_signal_type(self) -> bool:
        if self.signal_type in (SignalType.ST8T, SignalType.ST16T, SignalType.ST32T):
            return True

        return False

    def scale(self) -> int:
        if self.signal_type == SignalType.ST8T:
            return 2**7

        if self.signal_type == SignalType.ST16T:
            return 2**15

        if self.signal_type == SignalType.ST32T:
            return 2**31

        return 1

    def items_per_sample(self):
        if self.is_complex():
            return 2

        return 1



'''
Generate QAM signal with the given samples per symbol,
modulated with the provided constellation and symbol indices.
The signal is generated using a root raised cosine
filter with given beta.
'''
import numpy as np
from numpy import array as nparray
from scipy.signal import fftconvolve
from .root_raised_cosine_filter_generator import root_raised_cosine_filter_generator

def generate_qam(
    symbol_indices: nparray,
    constellation: nparray,
    samples_per_sym: int=2,
    beta: float=0.25,
) -> nparray:
    '''
    Parameters
    ----------
    symbol_indices:
        Symbol indices to be modulated.  Each index must be an integer at least 0
        and at most 1 less than the constellation size.
    constellation:
        Array of complex numbers describing the constellation points.
    samples_per_sym:
        Number of samples per symbol (Default 2).  Must be an integer at least 2.
    beta:
        Roll off factor (Default 0.25).  Must be a float between 0 and 1.

    Returns array of len(symbol_indices)*samples_per_sym complex samples.
    '''

    # Check Parameters
    # Symbol indices must be between 0 and length(constellation).
    if max(symbol_indices) < 0:
        raise ValueError("generate_qam error: All symbol indices must be at least 0")

    if max(symbol_indices) >= len(constellation):
        raise ValueError(f"generate_qam error: All symbol indices must be less than the constellation size ({len(constellation)})")

    # Constellation must have at least 2 points
    if len(constellation) < 2:
        raise ValueError(f"generate_qam error: Constellation size ({len(constellation)}) must be at least 2.")

    # If user provided sceale, then it must be nonnegative
    if beta < 0 or beta > 1:
        raise ValueError(f"generate_qam error: Beta ({beta}) must be between 0 and 1.")

    # Get to work

    # Create symbols from symbol indices
    symbols = constellation[symbol_indices]

    # Create pulse train
    pulse_train = np.kron(symbols, np.append([1], np.zeros(samples_per_sym-1)))

    # Create pulse shaping filter
    baud_width = 21
    pulse_shape = root_raised_cosine_filter_generator(baud_width, 1, samples_per_sym, beta)

    # Return
    return fftconvolve(pulse_train, pulse_shape, mode="same")



# vim: expandtab tabstop=4 shiftwidth=4

from enum import Enum

from .exceptions import EndiannessException

class Endianness(Enum):
    BIG = '>'
    LITTLE = '<'

def normalize_endianness(endianness: str) -> Endianness:
    endianness = endianness.strip().lower()

    if endianness in ('l', 'little'):
        return Endianness.LITTLE

    if endianness in ('b', 'big'):
        return Endianness.BIG

    raise EndiannessException(f"Unknown endianness {endianness}")



# vim: expandtab tabstop=4 shiftwidth=4

from numpy import pi, sin, cos, where, zeros
from numpy import array as nparray

def root_raised_cosine(time_array: nparray, spd: float, beta: float=0.25) -> nparray:
    '''
    Compute the values of the root raised cosine function at points in time_array.

    Parameters
    ----------
    time_array: nparray
        Input time values.
    spd: float
        Symbol period in seconds.  Reciprocal of the symbol rate.
    beta: float
        Roll off factor.  In the interval (0,1]
        Default is 0.25.

    Returns a numpy float array.
    '''

    # Define threshhold for zero
    zero_thresh = 10**-10

    # Create output array
    func_val = zeros(len(time_array))

    # Initial denominator values
    denominator = pi*time_array/spd*(1-16*(beta**2)*(time_array**2)/(spd**2))

    # Compute output at points where denominator1 is not zero
    nzidx = where(abs(denominator) >= zero_thresh)
    vals=time_array[nzidx]
    numerator1 = (sin(pi*vals/spd*(1-beta)) + 4*beta*vals/spd*cos(pi*vals/spd*(1+beta)))
    denominator1 = denominator[nzidx]
    func_val[nzidx] = numerator1/denominator1

    # Compute output at points where denominator1 is zero
    zidx = where(abs(denominator) < zero_thresh)
    vals=time_array[zidx]
    numerator2 = pi*(1-beta)*cos(pi*vals*(1-beta)/spd)/spd + 4*beta*cos(pi*vals*(1+beta)/spd)/spd - 4*pi*vals*beta*(1+beta)*sin(pi*vals*(1+beta)/spd)/(spd**2)
    denominator2 = pi*((spd**2)-48*(vals**2)*(beta**2))/(spd**3)
    func_val[zidx] = numerator2/denominator2

    return func_val

def rrcfunct(*args, **kwargs) -> nparray:
    '''
    Alias for root_raised_cosine.
    '''
    return root_raised_cosine(*args, **kwargs)



# vim: expandtab tabstop=4 shiftwidth=4

import numpy as np
from numpy import array as nparray

def vector_power(vector: nparray) -> nparray:
    """
    Compute the average power for a complex ndarray vector
    """

    # Equivalent Octave code: vector*vector'/length(vector)
    # Ensure vector is a complex row vector
    vector = np.array([vector.flatten()]).astype(complex)

    return ((vector.dot(np.transpose(vector.conjugate()))).real)[0, 0] / float(vector.shape[1])

def vecpow(*args, **kwargs) -> nparray:
    '''
    Alias for vector_power.
    '''
    return vector_power(*args, **kwargs)



# vim: expandtab tabstop=4 shiftwidth=4

from numpy import arange, kron, rint, ones
from numpy import array as nparray

def decimal_convert_to_base(decimal_number: nparray, base: int, num_digits: int):
    """
    Given an array of decimal integers, computes an array of numbers with given base
    Parameters
    ----------
    decimal_number:
        numpy.array of base 10 integers.
    base: int
        convert to integers using this base.
    num_digits: int
        number of digits output per input integer.

    Returns a numpy int array.
    """

    # Ensure we have integers
    decnum = rint(decimal_number).astype(int)
    # Exponent array
    exp_arr = kron(ones(decnum.shape, dtype=int), base**arange(num_digits-1, -1, -1, dtype=int))
    # Base array
    base_arr = kron(decnum, ones((1, num_digits), dtype=int))

    return (base_arr//exp_arr) % base

def num2base(*args, **kwargs) -> nparray:
    '''
    Alias for decimal_convert_to_base.
    '''
    return decimal_convert_to_base(*args, **kwargs)



# vim: expandtab tabstop=4 shiftwidth=4

class DSPFTWException(Exception):
    pass

class DataTypeException(DSPFTWException):
    pass

class EndiannessException(DSPFTWException):
    pass

class FileNameException(DSPFTWException):
    pass

class NumberSpaceException(DSPFTWException):
    pass

class SignalTypeException(DSPFTWException):
    pass

class WriteModeException(DSPFTWException):
    pass

class RoundingException(DSPFTWException):
    pass



# vim: expandtab tabstop=4 shiftwidth=4

import numpy as np

def true_zero(sigflat: np.array) -> np.array:
    return np.array(np.round(sigflat), 'int')

def true_one(sigflat: np.array) -> np.array:
    return np.array(sigflat, 'int')



'''
Generate QPSK signal from randomly generated symbols.
QPSK signal is complex basebanded and sampled at 'samples_per_sym'
samples per symbol.  The signal is generated using a root raised cosine
filter with given beta.
'''
import numpy as np
from numpy import exp, pi

from .exceptions import DSPFTWException
from .generate_qam import generate_qam
from .save_signal import save_signal
from .save_data import save_data

def generate_random_qpsk(ofile_samples: str,
                         ofile_symbols: str='',
                         sample_type: str='16t',
                         endianness: str='b',
                         num_symbols: int=10000,
                         samples_per_sym: int=2,
                         beta: float=0.25) -> int:
    '''
    Parameters
    ----------
    ofile_samples:
        Output file name for signal samples.
        Number of samples = num_symbols * samples_per_sym
    ofile_symbols:
        Output file name for signal samples.
        Will write symbols as int8
        Number of bits = 2 * num_symbols
        Number of bytes = 2 * num_symbols / 8
        Default: Empty string, which means do not write symbols to file.
    sample_type:
        The size and type of signal samples, such as '8t', '16t', '32t', '32f', '64f'.
        Default: 16t
    endianness:
        Endianness of the file, such as 'big', 'b', 'little', 'l'.
        Default: b
    num_symbols:
        Number of symbols to generate.
    samples_per_sym:
        Number of samples per symbol.  Must be an integer at least 2.
        Default: 2
    beta:
        Roll off factor.  Must be a float between 0 and 1.
        Default 0.25

    Returns number of samples written to file.
    '''

    # Check Parameters
    # Must have a positive number of symbols to generate
    if num_symbols <= 0:
        raise DSPFTWException('num_symbols must be positive')

    # Must have a positive number of samples per symbol
    if samples_per_sym <= 0:
        raise DSPFTWException('samples_per_sym must be positive')

    # Roll off factor must be between 0 and 1.
    if beta < 0 or beta > 1:
        raise DSPFTWException('Beta must be between 0 and 1')

    # Generate random symbols
    symbol_indices = np.random.randint(0, 2, num_symbols)

    # QPSK constellation
    constellation = exp(2j * pi * np.arange(1, 8, 2) / 8.0)

    # Generate QPSK signal
    signal = generate_qam(symbol_indices, constellation, samples_per_sym, beta)

    # Write signal to file
    num_written, _ = save_signal(ofile_samples, signal, sample_type, 'complex', endianness)

    # Optionally write symbol indices to file
    if ofile_symbols:
        save_data(ofile_symbols, symbol_indices, "int8")

    return num_written



# vim: expandtab tabstop=4 shiftwidth=4

from sys import byteorder

from numpy import array as nparray

from .data_types import FullDataType

def save_data(file_name: str, data: nparray, data_type: str, endianness: str=byteorder) -> int:
    '''
    Saves data to a file.

    Parameters
    ----------
    file_name: string
        The name of the file to save.
    data: numpy.array
        The data to save.
    data_type: string
        The type of data to save, such as 'uint16', 'int8', or 'float64'.
    endianness: string
        The endianness of the data, such as 'b', 'big', 'l', or 'little'.
        Defaults to the system byte order.

    Returns the number of elements saved.
    '''
    full_data_type = FullDataType(data_type, endianness)
    data.astype(full_data_type.numpy_dtype()).tofile(file_name)
    return data.size

def savedata(*args, **kwargs) -> nparray:
    '''
    Alias for save_data.
    '''
    return save_data(*args, **kwargs)



# vim: expandtab tabstop=4 shiftwidth=4

from enum import Enum
from typing import Any

from .exceptions import RoundingException
from .truncation import true_zero, true_one

class Rounding(Enum):
    TRUEZERO = 0
    TRUEONE = 1

def normalize_rounding(rounding: Any) -> Rounding:

    if isinstance(rounding, int):
        if rounding == 0:
            return Rounding.TRUEZERO
        if rounding == 1:
            return Rounding.TRUEONE

    if isinstance(rounding, str):
        if rounding in ('0', 'zero'):
            return Rounding.TRUEZERO
        if rounding in ('1', 'one'):
            return Rounding.TRUEONE

    if rounding is true_zero:
        return Rounding.TRUEZERO
    if rounding is true_one:
        return Rounding.TRUEONE

    raise RoundingException(f"Unknown rounding {rounding}")



# vim: expandtab tabstop=4 shiftwidth=4

from enum import Enum, auto
from os.path import splitext
from re import compile as regex_compile
from typing import Any, Dict, Tuple

from numpy import array as nparray

from .exceptions import FileNameException
from .load_bits import load_bits
from .load_signal import load_signal

signal_filename_pattern = regex_compile(r'\.(?P<number_space>real|cplx)'
                                        r'\.(?P<sample_rate>([0-9]+(\.[0-9]+)?))'
                                        r'\.(?P<extension>8t|16t|16tl|16tr|32t|32tl|32tr|32f|32fl|32fr|64f|64fl|64fr)')

class FileType(Enum):
    SIGNAL = auto()
    BITS = auto()

def get_file_type(filename: str) -> FileType:
    _, extension = splitext(filename)

    if extension in ('.8t',
                     '.16t', '.16tl', '.16tr',
                     '.32t', '.32tl', '.32tr',
                     '.32f', '.32fl', '.32fr',
                     '.64f', '.64fl', '.64fr'):
        return FileType.SIGNAL

    if extension == '.bits':
        return FileType.BITS

    raise FileNameException(f'Could not determine file type by extension {extension}')

def get_number_space(filename: str) -> str:
    if '.cplx.' in filename:
        return 'cplx'

    if '.real.' in filename:
        return 'real'

    raise FileNameException(f'Could not determine if file is real or complex based on filename {filename}')

def get_endianness(filename: str) -> str:
    _, extension = splitext(filename)

    if extension.endswith('l') or extension.endswith('r'):
        return 'little'

    return 'big'

def get_sample_rate(filename: str) -> float:
    match = signal_filename_pattern.search(filename)

    try:
        return float(match.group('sample_rate'))

    except AttributeError as ex:
        raise FileNameException(f'Could not determine sample rate from filename {filename}') from ex

def get_extension(filename: str) -> str:
    _, extension = splitext(filename)
    return extension[1:]  # remove leading dot

def filename_load_signal(filename: str, count: int, offset: int) -> Tuple[nparray, Dict[str, Any]]:
    signal = load_signal(filename,
                         get_extension(filename),
                         get_number_space(filename),
                         endianness=get_endianness(filename),
                         num_samples=count,
                         start_sample=offset)

    metadata = {
        'filename': filename,
        'sample_rate': get_sample_rate(filename),
    }

    return signal, metadata

def filename_load_bits(filename: str, count: int, offset: int) -> Tuple[nparray, Dict[str, Any]]:
    bits = load_bits(filename, count=count, offset=offset)

    metadata = {
        'filename': filename,
    }

    return bits, metadata

def filename_load(filename: str, count: int=-1, offset: int=0) -> Tuple[nparray, Dict[str, Any]]:
    '''
    Automatically loads a file based on the information in its filename.

    Parameters
    ----------
    filename:
        The file name to load.
    count:
        The number of items to load once the type is determined.
    offset:
        Number of items to skip once the type is determined.

    Returns a numpy array and a metadata dictionary.
    '''
    filename = str(filename)  # in case we get a pathlib.Path
    file_type = get_file_type(filename)

    if file_type == FileType.SIGNAL:
        return filename_load_signal(filename, count, offset)

    if file_type == FileType.BITS:
        return filename_load_bits(filename, count, offset)

    raise FileNameException(f'Could not determine how to load {filename}')

def fnload(*args, **kwargs) -> nparray:
    '''
    Alias for filename_load.
    '''
    return filename_load(*args, **kwargs)



# vim expandtab tabstop=4 shiftwidth=4

from numpy import pi, sinc, sin, cos, arange, where, zeros
from numpy import ceil  # pylint: disable=no-name-in-module
from numpy import array as nparray

def raised_cosine_filter_generator(symbol_width: int, symbol_rate: float, sample_rate: float, beta: float=0.25) -> nparray:
    '''
    Compute the taps of a raised cosine (RC) filter.

    Parameters
    ----------
    symbol_width: int
        Number of bauds wide for the output filter.
        An odd integer will ensure the initial and filanl taps are at baud boundaries.
    symbol_rate: float
        Symbol rate in Hz.
    sample_rate: float
        Sample rate in Hz.
    beta: float
        Roll off factor. In the interval (0,1]
        Default is 0.25.

    Returns a numply float array.
    '''

    # Define threshold for zero
    zero_thresh = 10**-10

    # Samples per baud
    spb = sample_rate / symbol_rate

    # Create the input time values
    time_array = arange(-ceil((symbol_width/2.0)*spb), ceil((symbol_width/2.0)*spb+1), dtype=float) / sample_rate

    # Create the filter values
    if beta == 0:
        filter_val = sinc(time_array)
    else:
        filter_val = zeros(time_array.shape)

        # Non-zero denominator
        nzidx = where(abs(1-(2*beta*time_array)**2) >= zero_thresh)
        inval = time_array[nzidx]
        filter_val[nzidx] = sinc(inval)*cos(pi*beta*inval)/(1-(2*beta*inval)**2)

        # Zero denominator: Using l'Hopital's rule
        zidx = where(abs(1-(2*beta*time_array)**2) < zero_thresh)
        inval = time_array[zidx]
        filter_val[zidx] = sinc(inval)*pi*sin(pi*beta*inval)/(8*beta*inval)

    return filter_val

def rcfiltgen(*args, **kwargs) -> nparray:
    '''
    Alias for raised_cosine_filter_generator.
    '''
    return raised_cosine_filter_generator(*args, **kwargs)



# vim: expandtab tabstop=4 shiftwidth=4

''' Cross-Correlation with normalization set to default
'''

from scipy.signal import convolve, correlate

import numpy as np

def signal_correlation(in1: np.array, in2: np.array, norm=True):
    '''
    Cross-correlates two arrays and normalizes the output
    Parameters
    ----------
    in1: 
        numpy.array containing first signal
    in2:
        numpy.array containing second signal
    norm: bool
        flag to normalize the output.
        default is True

    Returns the correlation vector as a numpy array.
    '''

    arr_l = in1
    arr_s = in2
    # If norm is set to False, unnormalized output will be returned
    out_arr = correlate(arr_l, arr_s, mode='full')
    # Default output is set to normalized
    if norm:
        s_pow = np.sqrt(np.sum(arr_s*arr_s.conj()))
        l_pow = np.sqrt(convolve(arr_l*arr_l.conj(), np.ones(len(arr_s)), mode='full'))*s_pow
        out_arr = out_arr/l_pow
    return out_arr

def sigcorr(*args, **kwargs):
    '''
    Alias for signal_correlation.
    '''
    return signal_correlation(*args, **kwargs)



# vim: expandtab tabstop=4 shiftwidth=4

from sys import byteorder

from numpy import array as nparray
from numpy import fromfile

from .data_types import FullDataType

def load_data(
    file_name: str,
    data_type: str,
    endianness: str=byteorder,
    count: int=-1,
    offset: int=0,
) -> nparray:
    '''
    Loads data from a file.

    Parameters
    ----------
    file_name: string
        The name of the file to load.
    data_type: string
        The type of data to load, such as 'uint16', 'int8', or 'float64'.
    endianness: string
        The endianness of the data, such as 'b', 'big', 'l', or 'little'.
        Defaults to the system byte order.
    count: int
        The number of data elements to load.  Defaults to the whole file.
    offset: int
        The number of initial data elements to skip in the file.  Defaults to zero.

    Returns a numpy array.
    '''
    full_data_type = FullDataType(data_type, endianness)
    data = fromfile(file_name,
                    dtype=full_data_type.numpy_dtype(),
                    offset=offset,
                    count=count)
    return data

def loaddata(*args, **kwargs) -> nparray:
    '''
    Alias for load_data.
    '''
    return load_data(*args, **kwargs)



# vim: expandtab tabstop=4 shiftwidth=4

from enum import Enum

from .endianness import Endianness, normalize_endianness
from .exceptions import DataTypeException

class DataType(Enum):
    U8   = 'u1'
    U16  = 'u2'
    U32  = 'u4'
    U64  = 'u8'
    I8   = 'i1'
    I16  = 'i2'
    I32  = 'i4'
    I64  = 'i8'
    F32  = 'f4'
    F64  = 'f8'
    F128 = 'f16'

def normalize_data_type(data_type: str) -> DataType:
    data_type = data_type.strip().lower()

    if data_type in ('u8', 'uint8'):
        return DataType.U8

    if data_type in ('u16', 'uint16'):
        return DataType.U16

    if data_type in ('u32', 'uint32'):
        return DataType.U32

    if data_type in ('u64', 'uint64'):
        return DataType.U64

    if data_type in ('i8', 'int8'):
        return DataType.I8

    if data_type in ('i16', 'int16'):
        return DataType.I16

    if data_type in ('i32', 'int32'):
        return DataType.I32

    if data_type in ('i64', 'int64'):
        return DataType.I64

    if data_type in ('f32', 'float32'):
        return DataType.F32

    if data_type in ('f64', 'float64'):
        return DataType.F64

    if data_type in ('f128', 'float128'):
        return DataType.F128

    raise DataTypeException(f"Unknown data type {data_type}")

class FullDataType:
    def __init__(self, data_type: str, endianness: str) -> None:
        self.data_type: DataType = normalize_data_type(data_type)
        self.endianness: Endianness = normalize_endianness(endianness)

    def numpy_dtype(self) -> str:
        return self.endianness.value + self.data_type.value



# vim: expandtab tabstop=4 shiftwidth=4

from enum import Enum

from .exceptions import WriteModeException

class WriteMode(Enum):
    WRITE = 'wb'
    APPEND = 'ab'

def normalize_write_mode(write_mode: str) -> WriteMode:
    write_mode = write_mode.strip().lower()

    if write_mode in ('w', 'write'):
        return WriteMode.WRITE

    if write_mode in ('a', 'append'):
        return WriteMode.APPEND

    raise WriteModeException(f"Unknown write mode {write_mode}")



'''
Reads QPSK signal file and associated symbol file, demodulates QPSK
to produce symbols and compares with symbols from symbol file
'''
import numpy as np
from scipy.signal import fftconvolve
from numpy import exp, pi

from .exceptions import DSPFTWException
from .load_signal import load_signal
from .load_data import load_data
from .root_raised_cosine_filter_generator import root_raised_cosine_filter_generator

def test_qpsk_modulation(ifile_samples: str,
                         ifile_symbols: str,
                         sample_type: str='16t',
                         endianness: str='b',
                         num_symbols: int=0,
                         samples_per_sym: int=2,
                         beta: float=0.25) -> int:
    '''
    Parameters
    ----------
    ifile_samples:
        Input file name for signal samples.
    ifile_symbols:
        Input file name for signal samples.
        Will read symbols as int8
    sample_type:
        The size and type of signal samples, such as '8t', '16t', '32t', '32f', '64f'.
        Default: 16t
    endianness:
        Endianness of the file, such as 'big', 'b', 'little', 'l'.
        Default: b
    num_symbols:
        Number of symbols to read.
        Will read 'num_symbols' symbols and num_symbols * samples_per_sym samples
        Default: 0 = read entire file
    samples_per_sym:
        Number of samples per symbol.  Must be an integer at least 2.
        Default: 2
    beta:
        Roll off factor.  Must be a float between 0 and 1.
        Default 0.25

    Returns number of symbol differences
    '''

    # Check Parameters
    # Must have a positive number of symbols to generate
    if num_symbols < 0:
        raise DSPFTWException('num_symbols must be nonnegative')

    # Must have a positive number of samples per symbol
    if samples_per_sym <= 0:
        raise DSPFTWException('samples_per_sym must be positive')

    # Roll off factor must be between 0 and 1.
    if beta < 0 or beta > 1:
        raise DSPFTWException('beta must be between 0 and 1')

    # Load symbols and signal
    if num_symbols == 0:
        symbol_indices = load_data(ifile_symbols, 'int8', 'b')
        signal = load_signal(ifile_samples, sample_type, 'complex', endianness)
    else:
        symbol_indices = load_data(ifile_symbols, 'int8', 'b', num_symbols)
        signal = load_signal(ifile_samples, sample_type, 'complex', endianness, num_symbols*samples_per_sym)

    # QPSK constellation as a row vector
    constellation = [exp(2j * pi * np.arange(1, 8, 2) / 8.0)]

    # Filter signal with root raised cosine filter to remove inter-symbol interference
    baud_width = 21
    rrc_filter = root_raised_cosine_filter_generator(baud_width, 1, samples_per_sym, beta)
    signal_eq = fftconvolve(signal, rrc_filter, mode='same')

    # Decimate by 'samples_per_symto get complex symbols
    xys = signal_eq[::samples_per_sym]

    # Demodulate to get symbol indices
    demod_indices = np.argmin(abs(xys - np.transpose(constellation)), 0).astype(int)

    # Compare symbol_indices and demod_indices
    if len(symbol_indices) != len(demod_indices):
        raise DSPFTWException(f"symbol_indices length != demod_indices length; {len(symbol_indices)} != {len(demod_indices)}")

    diff_count = 0

    for idx in np.arange(len(symbol_indices)):
        if symbol_indices[idx] != demod_indices[idx]:
            diff_count += 1

    return diff_count



# vim: expandtab tabstop=4 shiftwidth=4

from numpy import pi, sin, append, arange, zeros
from numpy import ceil  # pylint: disable=no-name-in-module
from numpy import array as nparray

from .root_raised_cosine import root_raised_cosine

def root_raised_cosine_filter_generator(symbol_width: int, symbol_rate: float, sample_rate: float, beta: float=0.25) -> nparray:
    '''
    Compute the taps of a root raised cosine (RRC) filter.

    Parameters
    ----------
    symbol_width: int
        Number of bauds wide for the output filter.
        An odd integer will ensure the initial and final taps are at baud boundaries.
    symbol_rate: float
        Symbol rate in Hz.
    sample_rate: float
        Sample rate in Hz.
    beta: float
        Roll off factor.  In the interval (0,1]
        Default is 0.25.

    Returns a numpy float array.
    '''

    # Set the symbol period
    spd = 1.0 / symbol_rate

    # Create the input time values
    time_array = arange(-ceil((symbol_width/2.0)*spd*sample_rate), ceil((symbol_width/2.0)*spd*sample_rate+1), dtype=float) / sample_rate

    # Create filter values
    if beta == 0:
        filter_val = zeros(time_array.shape)
        tlen = len(time_array)

        # Define filter everywhere except center tap
        idx = append(arange((tlen-1)//2, dtype=int), arange((tlen-1)//2+1, tlen, dtype=int))
        filter_val[idx] = sin(pi*time_array[idx]/spd) / (pi*time_array[idx]/spd)

        # Define filter at center tap
        filter_val[(tlen-1)//2] = 1

    else:
        filter_val = root_raised_cosine(time_array, spd, beta)

    return filter_val

def rrcfiltgen(*args, **kwargs) -> nparray:
    '''
    Alias for root_raised_cosine_filter_generator.
    '''
    return root_raised_cosine_filter_generator(*args, **kwargs)



# vim: expandtab tabstop=4 shiftwidth=4

from typing import IO
from numpy import array as nparray
from numpy import packbits

from .writemode import normalize_write_mode

def save_bits_to(writable: IO, bool_array: nparray, packed=True) -> int:
    '''
    Save bits to a writable object, such as an open file.
    This can be useful when you want to keep writing
    to the same file object, or reopen a file in append
    mode.

    Parameters
    ----------
    writable: file-like IO object
        A file-like object that has a write() method.
    bool_array: numpy.array
        The bool array.
    packed: bool
        Whether to pack the bits into bytes.
        Defaults to True.

    Returns the number of bytes written.

    Example
    -------
    # Opens an output file in append mode.
    with open('some_file.bits', 'a+b') as outfile:
        save_bits_to(outfile, array_of_bits)
    '''

    if packed:
        bytes_to_write = packbits(bool_array).tobytes()  # pylint: disable=E1101

    else:
        bytes_to_write = nparray(bool_array).tobytes()

    writable.write(bytes_to_write)
    return len(bytes_to_write)

def save_bits(file_name: str, bool_array: nparray, packed=True, write_mode: str='w') -> int:
    '''
    Save bits to a file from a bool array.

    Parameters
    ----------
    file_name: string
        The name of the file to save.
    bool_array: numpy.array
        The bool array.
    packed: bool
        Whether to pack the bits into bytes.
        Default is True.
    write_mode: str
        'w' for new or overwrite. 'a' for append.
        Default is w
    Returns the number of bytes written.
    '''

    write_mode = normalize_write_mode(write_mode)

    if len(bool_array) == 0:
        bool_array = nparray([], dtype=int)

    with open(file_name, write_mode.value) as bit_file:
        return save_bits_to(bit_file, bool_array, packed)

def savebits(*args, **kwargs) -> int:
    '''
    Alias for save_bits.
    '''
    return save_bits(*args, **kwargs)
